//Illinois Open Source License
//
//University of Illinois
//Open Source License
//
//Copyright © 2014,    Board of Trustees of the University of Illinois.  All rights reserved.
//
//Developed by:
//
// Distributed Protocols Research Group in the Department of Computer Science
// The University of Illinois at Urbana-Champaign
// http://dprg.cs.uiuc.edu/

//Imranul Hoque, ihoque2@illinois.edu
//Indranil Gupta, indy@cs.uiuc.edu
//
//Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal with the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimers.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimers in the documentation and/or other materials provided with the distribution.
//    * Neither the names of The Monet Group or The University of Illinois at Urbana-Champaign, nor the names of its contributors may be used to endorse or promote products derived from this Software without specific prior written permission.
//
//THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
//PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
//AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.


#include "barrier.h"

using namespace std;

NewBarrier::NewBarrier(const char *address, int port) {
	strcpy(this->address, address);
	this->port = port;
}

bool NewBarrier::enter(int active_count) {
	boost::asio::io_service io_service;

	tcp::resolver resolver(io_service);
	char s_port[16];
	sprintf(s_port, "%d", port);
	tcp::resolver::query query(this->address, s_port);
	tcp::resolver::iterator endpoint_iterator = resolver.resolve(query);
	tcp::resolver::iterator end;

	tcp::socket socket(io_service);
	boost::system::error_code error = boost::asio::error::host_not_found;
	while (error && endpoint_iterator != end) {
		socket.close();
		socket.connect(*endpoint_iterator++, error);
	}
	if (error)
		throw boost::system::system_error(error);
	int t_active = 0;
	boost::asio::write(socket, boost::asio::buffer(&active_count, sizeof(int)),
				boost::asio::transfer_all(), error);
	boost::asio::read(socket, boost::asio::buffer(&t_active, sizeof(int)));
	if(t_active == 0) {
		return false;
	}
	return true;
}
/*
zhandle_t *Barrier::zh = NULL;
Barrier::Barrier(const char *address, const char *root, const char *name,
		int size) {
	if (zh == NULL) {
		zh = zookeeper_init(address, NULL, 10000, 0, 0, 0);
	}
	zoo_set_debug_level((ZooLogLevel) 0);
	struct Stat stat;
	char buffer[128];
	strcpy(this->root, "/");
	strcat(this->root, root);
	strcpy(this->name, "/");
	strcat(this->name, name);
	this->size = size;

	int rc = zoo_exists(zh, root, 0, &stat);
	if (rc != ZOK) {
		zoo_create(zh, this->root, "dummy", 5, &ZOO_OPEN_ACL_UNSAFE, 0, buffer,
				sizeof(buffer) - 1);
	}

}

void Barrier::cleanup() {
	if (zh) {
		zookeeper_close(zh);
	}
}

bool Barrier::enter() {
	boost::mutex::scoped_lock lock(io_mutex);
	char child[128];
	char buffer[128];
	sprintf(child, "%s%s", root, name);
	zoo_create(zh, child, "value", 5, &ZOO_OPEN_ACL_UNSAFE,
			ZOO_EPHEMERAL | ZOO_SEQUENCE, buffer, sizeof(buffer) - 1);
	struct String_vector children;
	while (true) {
		zoo_wget_children(zh, root, watcher, this, &children);
		if (children.count < size) {
			cond.wait(lock);
		} else {
			break;
		}
	}
	return true;
}
Barrier::~Barrier() {
	//sleep(1);
	//zookeeper_close(zh);
}

void watcher(zhandle_t *zzh, int type, int state, const char *path,
		void *watcherCtx) {
	Barrier *b = (Barrier *) watcherCtx;
	b->cond.notify_all();
}
*/
