//Illinois Open Source License
//
//University of Illinois
//Open Source License
//
//Copyright © 2014,    Board of Trustees of the University of Illinois.  All rights reserved.
//
//Developed by:
//
// Distributed Protocols Research Group in the Department of Computer Science
// The University of Illinois at Urbana-Champaign
// http://dprg.cs.uiuc.edu/

//Imranul Hoque, ihoque2@illinois.edu
//Indranil Gupta, indy@cs.uiuc.edu
//
//Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal with the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimers.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimers in the documentation and/or other materials provided with the distribution.
//    * Neither the names of The Monet Group or The University of Illinois at Urbana-Champaign, nor the names of its contributors may be used to endorse or promote products derived from this Software without specific prior written permission.
//
//THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
//PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
//AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.


#include "langar.h"

void Langar::session(socket_ptr sock) {
	if (!optimized) {
		try {
			int remote_host_id;
			boost::asio::read(*sock,
					boost::asio::buffer(&remote_host_id, sizeof(int)));
			int array_length = out_host[remote_host_id].index.getSize();
			double *data = new double[array_length];

			for (int i = 0; i < out_host[remote_host_id].index.getSize(); i++) {
				int gid = out_host[remote_host_id].index.data[i];
				int pid = gid_to_pid(gid);
				int lid = gid_to_lid(gid);
				data[i] = v[pid][lid].value;
			}

			boost::asio::write(*sock,
					boost::asio::buffer(data, array_length * sizeof(double)));
			sock->close();
			delete[] data;
		} catch (std::exception& e) {
			std::cerr << "Exception in thread: " << e.what() << "\n";
		}
	} else {
		try {
			int remote_host_id;

			boost::asio::read(*sock,
					boost::asio::buffer(&remote_host_id, sizeof(int)));

			int *index = new int[out_host[remote_host_id].index.getSize()];
			int array_length = 0;
			//vdata.reserve(host[remote_host_id].index.getSize());
			double *data = new double[out_host[remote_host_id].index.getSize()];
			for (int i = 0; i < out_host[remote_host_id].index.getSize(); i++) {
				int gid = out_host[remote_host_id].index.data[i];
				int pid = gid_to_pid(gid);
				int lid = gid_to_lid(gid);
				if (v[pid][lid].updated) { //(updated[gid]) {
					data[array_length] = v[pid][lid].value;
					index[array_length++] = gid;
				}
			}

			//double *data = new double[array_length];

			boost::asio::write(*sock,
					boost::asio::buffer(&array_length, sizeof(int)));
			boost::asio::write(*sock,
					boost::asio::buffer(index, sizeof(int) * array_length));
			/*
			 for (int i = 0; i < array_length; i++) {

			 int pid = gid_to_pid(index[i]);
			 int lid = gid_to_lid(index[i]);
			 data[i] = v[pid][lid].value;
			 //data[i] = write_buffer[index[i]];
			 }*/
			boost::asio::write(*sock,
					boost::asio::buffer(data, array_length * sizeof(double)));
			sock->close();
			delete[] data;
			delete[] index;
		} catch (std::exception& e) {
			std::cerr << "Exception in thread: " << e.what() << "\n";
		}
	}

}

int Langar::gid_to_lid(int id) {
	return (id / (thread_count * machine_count));
}

int Langar::gid_to_pid(int id) {
	return (id / machine_count) % thread_count;
}

void Langar::optimized_session(socket_ptr sock) {
	try {
		int array_length;
		int remote_machine_id;
		boost::asio::read(*sock,
				boost::asio::buffer(&array_length, sizeof(int)));
		boost::asio::read(*sock,
				boost::asio::buffer(&remote_machine_id, sizeof(int)));
		int *index = new int[array_length];

		boost::asio::read(*sock,
				boost::asio::buffer(index, sizeof(int) * array_length));

		for (int i = 0; i < array_length; i++) {
			//int pid = gid_to_pid(index[i]);
			//int lid = gid_to_lid(index[i]);
			out_host[remote_machine_id].index.add(index[i]);
			//v[pid][lid].p_list.push_back(remote_machine_id);
		}
		boost::asio::write(*sock,
				boost::asio::buffer(&array_length, sizeof(int)));
		//cout << remote_machine_id << "->" << out_host[remote_machine_id].index.getSize() << endl;
		sock->close();
		delete[] index;
	} catch (std::exception& e) {
		std::cerr << "Exception in thread: " << e.what() << "\n";
	}

}

void Langar::server() {
	boost::asio::io_service io_service;
	tcp::acceptor a(io_service,
			tcp::endpoint(tcp::v4(), host[machine_id].getPort()));
	while (running) {
		socket_ptr sock(new tcp::socket(io_service));
		a.accept(*sock);
		boost::thread t(boost::bind(&Langar::session, this, sock));
	}
}

void Langar::optimized_server() {
	boost::asio::io_service io_service;
	tcp::acceptor a(io_service,
			tcp::endpoint(tcp::v4(), host[machine_id].getPort() + 1));
	int count = 0;
	while (count < machine_count - 1) {
		socket_ptr sock(new tcp::socket(io_service));
		a.accept(*sock);
		boost::thread t(boost::bind(&Langar::optimized_session, this, sock));
		count++;
	}
}

void Langar::client(int host_id) {

	boost::asio::io_service io_service;

	tcp::resolver resolver(io_service);
	char s_port[16];
	sprintf(s_port, "%d", host[host_id].getPort());
	tcp::resolver::query query(host[host_id].getAddress(), s_port);
	tcp::resolver::iterator endpoint_iterator = resolver.resolve(query);
	tcp::resolver::iterator end;

	tcp::socket socket(io_service);
	boost::system::error_code error = boost::asio::error::host_not_found;
	while (error && endpoint_iterator != end) {
		socket.close();
		socket.connect(*endpoint_iterator++, error);
	}
	if (error)
		throw boost::system::system_error(error);

	if (!optimized) {
		int array_length = host[host_id].index.getSize();
		boost::asio::write(socket,
				boost::asio::buffer(&machine_id, sizeof(int)),
				boost::asio::transfer_all(), error);

		boost::array<double, 32768> buf;
		int l = 0;
		int batch_size =
				(int) buf.size() < array_length ? buf.size() : array_length;
		int remaining_size = array_length;
		while (true) {
			boost::system::error_code error;
			boost::asio::read(socket,
					boost::asio::buffer(buf, sizeof(double) * batch_size));
			for (int i = 0; i < batch_size; i++) {
				int key = host[host_id].index.data[l];
				read_buffer[key] = buf[i];
				l++;
			}
			remaining_size -= batch_size;
			if (remaining_size == 0) {
				break;
			}
			if (remaining_size < batch_size) {
				batch_size = remaining_size;
			}
			//cout << endl;
			if (error == boost::asio::error::eof) {
				break; // Connection closed cleanly by peer.
			} else if (error) {
				throw boost::system::system_error(error); // Some other error.
			}
		}

	} else {
		int array_length;
		boost::asio::write(socket,
				boost::asio::buffer(&machine_id, sizeof(int)),
				boost::asio::transfer_all(), error);
		boost::asio::read(socket,
				boost::asio::buffer(&array_length, sizeof(int)));
		int *index = new int[array_length];
		boost::asio::read(socket,
				boost::asio::buffer(index, sizeof(int) * array_length));

		boost::array<double, 32768> buf;
		int l = 0;
		int batch_size =
				(int) buf.size() < array_length ? buf.size() : array_length;
		int remaining_size = array_length;
		while (true) {
			boost::system::error_code error;
			boost::asio::read(socket,
					boost::asio::buffer(buf, sizeof(double) * batch_size));
			for (int i = 0; i < batch_size; i++) {
				int key = index[l];
				read_buffer[key] = buf[i];
				updated[key] = true;
				//cout << "Received information about: " << key << endl;
				/*
				 for (int j = 0; j < in_list[key].size(); j++) {
				 int gid = in_list[key][j];
				 //cout << "Setting " << gid << " as active." << endl;
				 int lid = gid_to_lid(gid);
				 int pid = gid_to_pid(gid);
				 v[pid][lid].active = true;
				 }
				 */
				////cout << endl;
				l++;
			}
			remaining_size -= batch_size;
			if (remaining_size == 0) {
				break;
			}
			if (remaining_size < batch_size) {
				batch_size = remaining_size;
			}
			//cout << endl;
			if (error == boost::asio::error::eof) {
				break; // Connection closed cleanly by peer.
			} else if (error) {
				throw boost::system::system_error(error); // Some other error.

			}
		}
		delete[] index;
	}
}

void Langar::optimized_client(int host_id) {
	boost::asio::io_service io_service;

	tcp::resolver resolver(io_service);
	char s_port[16];
	sprintf(s_port, "%d", host[host_id].getPort() + 1);
	tcp::resolver::query query(host[host_id].getAddress(), s_port);
	tcp::resolver::iterator endpoint_iterator = resolver.resolve(query);
	tcp::resolver::iterator end;

	tcp::socket socket(io_service);
	boost::system::error_code error = boost::asio::error::host_not_found;
	while (error && endpoint_iterator != end) {
		socket.close();
		socket.connect(*endpoint_iterator++, error);
	}
	if (error)
		throw boost::system::system_error(error);
	int array_length = host[host_id].index.getSize();
	//cout << "Local: " << machine_id << ", Remote: " << host_id << " | Count: " << array_length << endl;
	boost::asio::write(socket, boost::asio::buffer(&array_length, sizeof(int)),
			boost::asio::transfer_all(), error);
	boost::asio::write(socket, boost::asio::buffer(&machine_id, sizeof(int)),
			boost::asio::transfer_all(), error);
	boost::asio::write(socket,
			boost::asio::buffer(host[host_id].index.data,
					array_length * sizeof(int)), boost::asio::transfer_all(),
			error);
	boost::asio::read(socket, boost::asio::buffer(&array_length, sizeof(int)));
	/*
	 boost::array<double, 10000> buf;
	 int l = 0;
	 while (true) {
	 boost::system::error_code error;
	 size_t len = socket.read_some(boost::asio::buffer(buf), error);
	 for (int i = 0; i < len / sizeof(double); i++) {
	 read_buffer[host[host_id].index.data[l]] = buf[i];
	 l++;
	 }
	 if (error == boost::asio::error::eof) {
	 break; // Connection closed cleanly by peer.
	 } else if (error)
	 throw boost::system::system_error(error); // Some other error.
	 } */
}

Langar::Langar(int machine_id, const char *filename, int machine_count,
		int thread_count, int vertex_count, Host *host, const char *zkserver,
		int zkport, bool optimized = true) {
	this->machine_id = machine_id;
	if (filename != NULL) {
		strcpy(this->filename, filename);
		this->synthetic = false;
	} else {
		cout << "Synthetic Graph" << endl;
		this->synthetic = true;
	}
	this->machine_count = machine_count;
	this->vertex_count = vertex_count;
	this->thread_count = thread_count;
	this->host = new Host[machine_count];
	this->out_host = new Host[machine_count];
	this->barrier = new NewBarrier(zkserver, zkport);
	strcpy(this->zkserver, zkserver);
	this->zkport = zkport;
	this->optimized = optimized;
	this->running = true;
	this->updated = new bool[vertex_count];
	this->read_buffer = new double[vertex_count];
	this->v = NULL;
	this->partition_vcount = NULL;
	this->active_count = new int[thread_count];
	for (int i = 0; i < machine_count; i++) {
		this->host[i] = host[i];
		this->out_host[i] = host[i];
	}

	boost::thread t_o(boost::bind(&Langar::optimized_server, this));
	boost::thread t(boost::bind(&Langar::server, this));
	if (!synthetic) {
		load_vertex();
	} else {
		load_lognormal();
	}
	cout << "Done Loading" << endl;
	barrier->enter();

	boost::thread_group tgroup;
	for (int i = 0; i < machine_count; i++) {
		if (i == machine_id)
			continue;
		tgroup.create_thread(boost::bind(&Langar::optimized_client, this, i));
	}
	tgroup.join_all();

	barrier->enter();
	cout << "Starting Compute" << endl;
	timer.start();
	if (optimized) {
		optimized_compute();
	} else {
		compute();
	}
	timer.stop();
	timer.print();
	this->running = false;
	t_o.interrupt();
	t.interrupt();

}

void Langar::load(int wid, vector<long long> &degree, bool *flag) {

	typedef boost::uniform_int<long long> LongDistribution;
	typedef boost::mt19937_64 RandomNumberGenerator;
	typedef boost::variate_generator<RandomNumberGenerator&, LongDistribution> LongGenerator;

	RandomNumberGenerator generator;
	LongDistribution long_distribution(0, degree[vertex_count - 1] - 1);
	LongGenerator longGenerator(generator, long_distribution);

	generator.seed(thread_count * machine_id + wid);

	int average_degree = degree[vertex_count - 1] / vertex_count;

	partition_vcount[wid] = (vertex_count / thread_count) / machine_count;
	v[wid] = new Vertex[partition_vcount[wid]];
	for (int j = 0; j < partition_vcount[wid]; j++) {
		v[wid][j].id = j * thread_count * machine_count + wid * machine_count
				+ machine_id;
		v[wid][j].fcount =
				v[wid][j].id % 10 == 0 ? average_degree + 1 : average_degree;
		v[wid][j].f_list = new int[v[wid][j].fcount];

		for (int k = 0; k < v[wid][j].fcount;) {
			long long r = longGenerator();
			int fid = std::upper_bound(degree.begin(), degree.end(), r)
					- degree.begin();
			if (fid == v[wid][j].id)
				continue;
			v[wid][j].f_list[k++] = fid;

			if (fid % machine_count != machine_id) {
				flag[fid] = true;
				//if (!flag[fid]) {
				//host[fid % machine_count].index.add(fid);
				//flag[fid] = true;
				//}
			}
		}

	}
}

void Langar::load_lognormal() {

	typedef boost::normal_distribution<> NormalDistribution;
	typedef boost::uniform_int<long long> LongDistribution;
	typedef boost::mt19937_64 RandomNumberGenerator;
	typedef boost::variate_generator<RandomNumberGenerator&, NormalDistribution> NormalGenerator;

	NormalDistribution normal_distribution;

	RandomNumberGenerator generator;

	NormalGenerator normalGenerator(generator, normal_distribution);

	generator.seed(10);

	vector<long long> degree(vertex_count);

	for (int i = 0; i < vertex_count; i++) {
		degree[i] = (i > 0 ? degree[i - 1] : 0)
				+ (int) (exp(4 + 1.3 * normalGenerator()) + 0.5);
	}
	cout << degree[vertex_count - 1] << endl;

	v = new Vertex *[thread_count];
	partition_vcount = new int[thread_count];

	bool *flag = new bool[vertex_count];
	memset(flag, 0, vertex_count * sizeof(bool));

	boost::threadpool::pool tpool(thread_count);

	cout << "Start Loading Vertices." << endl;

	for (int i = 0; i < thread_count; i++) {
		tpool.schedule(
				boost::bind(&Langar::load, this, i, degree, flag));
	}
	tpool.wait();

	for(int i = 0 ; i < vertex_count ; i++) {
		if(flag[i]) {
			host[i % machine_count].index.add(i);
		}
	}
	delete[] flag;
	exit(0);
	/*
	 boost::asio::io_service io_service;
	 tcp::resolver resolver(io_service);

	 char s_port[16];
	 sprintf(s_port, "%d", zkport + 1);
	 tcp::resolver::query query(zkserver, s_port);
	 tcp::resolver::iterator endpoint_iterator = resolver.resolve(query);
	 tcp::resolver::iterator end;

	 tcp::socket socket(io_service);
	 boost::system::error_code error = boost::asio::error::host_not_found;
	 while (error && endpoint_iterator != end) {
	 socket.close();
	 socket.connect(*endpoint_iterator++, error);
	 }
	 if (error)
	 throw boost::system::system_error(error);

	 boost::asio::write(socket, boost::asio::buffer(&machine_id, sizeof(char)),
	 boost::asio::transfer_all(), error);

	 int *degree = new int[vertex_count / machine_count];

	 boost::asio::read(socket, boost::asio::buffer(degree, sizeof(int)
	 * vertex_count / machine_count));

	 v = new Vertex *[thread_count];
	 partition_vcount = new int[thread_count];

	 boost::dynamic_bitset<> flag(vertex_count);
	 flag.reset();

	 cout << "Start Loading Vertices" << endl;
	 int fid = machine_id;
	 for (int i = 0; i < thread_count; i++) {
	 partition_vcount[i] = (vertex_count / thread_count) / machine_count;
	 v[i] = new Vertex[partition_vcount[i]];
	 for (int j = 0; j < partition_vcount[i]; j++) {
	 v[i][j].id = j * thread_count * machine_count + i * machine_count
	 + machine_id;
	 v[i][j].fcount = min(degree[v[i][j].id / machine_count],
	 vertex_count - 1);
	 v[i][j].f_list = new int[v[i][j].fcount];

	 for (int k = 0; k < v[i][j].fcount;) {
	 fid = (fid + 2654435761U) % vertex_count;//rand() % vertex_count;
	 if (fid == v[i][j].id)
	 continue;
	 v[i][j].f_list[k++] = fid;

	 if (fid % machine_count != machine_id) {
	 if (!flag[fid]) {
	 host[fid % machine_count].index.add(fid);
	 flag[fid] = true;
	 }
	 }
	 }

	 }
	 }
	 for (int i = 0; i < machine_count; i++) {
	 if (i == machine_id)
	 continue;
	 cout << "Local: " << machine_id << ", Remote: " << i << " | Count: "
	 << host[i].index.getSize() << endl;
	 }
	 */
}

void Langar::load_vertex() {
	ifstream fin(filename);
	assert(fin.is_open());
	int batch_size = 10000;

	int *alloc_size = new int[thread_count];

	v = new Vertex *[thread_count];
	partition_vcount = new int[thread_count];
	bool *flag = new bool[vertex_count];
	memset(flag, false, vertex_count * sizeof(bool));

	for (int i = 0; i < thread_count; i++) {
		v[i] = (Vertex *) malloc(batch_size * sizeof(Vertex));
		partition_vcount[i] = 0;
		alloc_size[i] = batch_size;
	}

	for (int i = 0;; i++) {
		int id, fcount, fid;
		fin >> id;
		if (fin.eof()) {
			break;
		}

		int pid = (id / machine_count) % thread_count;

		if (partition_vcount[pid] >= alloc_size[pid]) {
			alloc_size[pid] += batch_size;
			Vertex *new_v = (Vertex *) realloc(v[pid],
					sizeof(Vertex) * alloc_size[pid]);
			assert(new_v != NULL);
			v[pid] = new_v;
		}

		assert(partition_vcount[pid] == id/(thread_count * machine_count));
		v[pid][partition_vcount[pid]].id = id;

		fin >> fcount;
		v[pid][partition_vcount[pid]].fcount = fcount;

		v[pid][partition_vcount[pid]].f_list = new int[fcount];
		int count = 0; //TODO
		for (int j = 0; j < fcount; j++) {
			fin >> fid;
			if (fid == id) { //TODO
				count++;
				continue;
			}
			v[pid][partition_vcount[pid]].f_list[j - count] = fid;
			//if (optimized) {
			//	in_list[fid].push_back(id);
			//}

			if (fid % machine_count != machine_id) {
				if (!flag[fid]) {
					host[fid % machine_count].index.add(fid);
					flag[fid] = true;
				}
			}
		}
		v[pid][partition_vcount[pid]].fcount = fcount - count;

		partition_vcount[pid]++;

	}

	for (int i = 0; i < machine_count; i++) {
		if (i == machine_id)
			continue;
		sort(host[i].index.data, host[i].index.data + host[i].index.getSize());
		cout << "Local: " << machine_id << ", Remote: " << i << " | Count: "
				<< host[i].index.getSize() << endl;

	}
	delete[] flag;
	delete[] alloc_size;
}

void Langar::memcopy(int wid) {
	for (int i = 0; i < partition_vcount[wid]; i++) {
		if (optimized) {
			updated[v[wid][i].id] = v[wid][i].updated;
		}
		//read_buffer[v[wid][i].id] = write_buffer[v[wid][i].id];
		read_buffer[v[wid][i].id] = v[wid][i].value;
	}
}

void Langar::optimized_compute() {
	//cout << "Inside optimized_compute()" << endl;
	//boost::thread_group tgroup;
	boost::threadpool::pool tpool(boost::thread::hardware_concurrency() - 2);
	for (int step = 0;; step++) {
		//cout << "Step: " << step << endl;

		for (int i = 0; i < thread_count; i++) {
			tpool.schedule(
					boost::bind(&Langar::optimized_worker, this, i, step));
			//tgroup.create_thread(boost::bind(&Langar::optimized_worker, this,
			//		i, step));
		}
		//tgroup.join_all();
		tpool.wait();
		int t_active = 0;
		for (int i = 0; i < thread_count; i++) {
			t_active += active_count[i];
		}
		memset(updated, false, vertex_count);

		for (int i = 0; i < thread_count; i++) {
			tpool.schedule(boost::bind(&Langar::memcopy, this, i));
		}

		if (!barrier->enter(t_active))
			break;

		for (int i = machine_id + 1; i < machine_count + machine_id; i++) {
			tpool.schedule(
					boost::bind(&Langar::client, this, i % machine_count));
		}

		tpool.wait();
		barrier->enter();

	}
	/*
	 for (int i = 0; i < partition_vcount[0]; i++) {
	 cout << v[0][i].id << ": " << write_buffer[v[0][i].id] << ", " << read_buffer[v[0][i].id] << endl;
	 }
	 */
	//cout << "Done compute." << endl;
}

void Langar::compute() {

	//boost::thread_group tgroup;
	boost::threadpool::pool tpool(boost::thread::hardware_concurrency() - 3);

	for (int step = 0; step < 5; step++) {
		//cout << "Iteration " << step << endl;
		for (int i = 0; i < thread_count; i++) {
			//tgroup.create_thread(boost::bind(&Langar::worker, this, i, step));
			//tgroup.create_thread(boost::bind(&Langar::pagerank, this, i, step));
			tpool.schedule(boost::bind(&Langar::pagerank, this, i, step));
		}
		//tgroup.join_all();
		tpool.wait();
		//cout << "Finished Itderation" << endl;

		for (int i = 0; i < thread_count; i++) {
			//tgroup.create_thread(boost::bind(&Langar::memcopy, this, i));
			tpool.schedule(boost::bind(&Langar::memcopy, this, i));
		}

		barrier->enter();
		if (step == 4)
			break;
		for (int i = machine_id + 1; i < machine_id + machine_count; i++) {
			tpool.schedule(
					boost::bind(&Langar::client, this, i % machine_count));
			//tgroup.create_thread(boost::bind(&Langar::client, this, i % machine_count));
			//if(count % 4 == 0) tg.join_all();
		}
		//tg.join_all();
		//tgroup.join_all();
		tpool.wait();
		barrier->enter();
	}
	/*
	 for (int i = 0; i < partition_vcount[0]; i++) {
	 cout << v[0][i].id << ": " << write_buffer[v[0][i].id] << endl;
	 }
	 */
	//cout << "Done compute." << endl;
}

void Langar::optimized_worker(int wid, int step) {
	int i, j;
	active_count[wid] = 0;
	if (step > 0) {
		for (i = 0; i < partition_vcount[wid]; i++) {
			v[wid][i].updated = false;
			int friend_count = v[wid][i].fcount;
			int min = INT_MAX;
			for (j = 0; j < friend_count; j++) {
				int fid = v[wid][i].f_list[j];
				if (!updated[fid]) {
					continue;
				}
				if (read_buffer[fid] < min) {
					min = read_buffer[fid];
				}
			}

			if (read_buffer[v[wid][i].id] - 1 > min) {
				v[wid][i].updated = true;
				active_count[wid]++;
				v[wid][i].value = min + 1;
				//write_buffer[v[wid][i].id] = min + 1;
			}

		}

	} else {
		for (i = 0; i < partition_vcount[wid]; i++) {
			if (v[wid][i].id == 0) {
				active_count[wid]++;
				//cout << "Activating single node" << endl;
				v[wid][i].updated = true;
				v[wid][i].value = 0;
				//write_buffer[v[wid][i].id] = 0;

			} else {
				v[wid][i].updated = false;
				v[wid][i].value = INT_MAX;
				//write_buffer[v[wid][i].id] = INT_MAX;
			}
		}

	}
}

void Langar::shortest_path(int wid, int step) {
	int i, j;
	if (step > 0) {
		for (i = 0; i < partition_vcount[wid]; i++) {
			int friend_count = v[wid][i].fcount;
			int min = INT_MAX;
			for (j = 0; j < friend_count; j++) {

				int fid = v[wid][i].f_list[j];
				if (read_buffer[fid] < min) {
					min = read_buffer[fid];
				}
			}

			if ((int) read_buffer[v[wid][i].id] - 1 > min) {
				v[wid][i].value = min + 1;
				//write_buffer[v[wid][i].id] = min + 1;
			}
		}
	} else {
		for (i = 0; i < partition_vcount[wid]; i++) {
			if (v[wid][i].id == 0) {
				v[wid][i].value = 0;
				//write_buffer[v[wid][i].id] = 0;

			} else {
				v[wid][i].value = INT_MAX;
				//write_buffer[v[wid][i].id] = INT_MAX;

			}
		}
	}
}
/*
 void Langar::optimized_worker(int wid, int step) {
 int i, j;

 if (step > 0) {
 for (i = 0; i < partition_vcount[wid]; i++) {
 if(!v[wid][i].active) continue;
 int friend_count = v[wid][i].fcount;
 double sum = 0;
 for (j = 0; j < friend_count; j++) {

 int fid = v[wid][i].f_list[j];
 sum += read_buffer[fid];
 }
 sum = 0.15 + 0.85 * sum;
 write_buffer[v[wid][i].id] = sum / friend_count;
 double error = abs(write_buffer[v[wid][i].id] - read_buffer[v[wid][i].id]);
 if(error < 1E-5) {
 v[wid][i].active = false;
 }

 }
 } else {
 for (i = 0; i < partition_vcount[wid]; i++) {
 write_buffer[v[wid][i].id] = 1;
 if(v[wid][i].id == 0) {
 v[wid][i].active = true;
 } else {
 v[wid][i].active = false;
 }
 }
 }

 }
 */

void Langar::pagerank(int wid, int step) {
	int i, j;
	if (step > 0) {

		for (i = 0; i < partition_vcount[wid]; i++) {
			int friend_count = v[wid][i].fcount;
			double sum = 0;
			for (j = 0; j < friend_count; j++) {

				int fid = v[wid][i].f_list[j];
				sum += read_buffer[fid];
			}
			sum = 0.15 + 0.85 * sum;
			v[wid][i].value = sum / friend_count;
			//write_buffer[v[wid][i].id] = sum / friend_count;
		}
	} else {
		for (i = 0; i < partition_vcount[wid]; i++) {
			//write_buffer[v[wid][i].id] = 1;
			v[wid][i].value = 1;
		}
	}
}

Langar::~Langar() {

	//Barrier::cleanup();
	for (int i = 0; i < thread_count; i++) {
		if (v) {
			if (v[i]) {
				for (int j = 0; j < partition_vcount[i]; j++) {
					delete[] v[i][j].f_list;
				}
				free(v[i]);
			}
		}
	}
	if (active_count)
		delete[] active_count;
	if (v)
		delete[] v;
	if (host)
		delete[] host;
	if (partition_vcount)
		delete[] partition_vcount;
	//if (write_buffer)
	//	delete[] write_buffer;
	if (read_buffer)
		delete[] read_buffer;
	if (updated)
		delete[] updated;
	if (out_host)
		delete[] out_host;
	if (barrier)
		delete barrier;
}
